/// Smart-WiFly-web server is a very primitive server using WiFly.
/// 
/// Working version, but it is not using the standardized wifly
/// library, which would not work for me... I had to make a number
/// of changes to get it to work well.
/// 
/// @caution If nothing visits its web page for a long time, it seems
/// to disable the interface. Perhaps a sleep timer I'm not spotting?
///
/// I created this because I couldn't fine one that worked and wanted to
/// learn the WiFly module. There are a lot of possible improvements:
/// @li I think I'm not using the Socket interface as fully as I should.
/// @li I would like it to be non-blocking.
/// @li I would like it to be faster (the interface from mbed to wifly is 
///     limited to 230400 baud before it drops chars.
/// @li I would like to integrate this with the rtos.
///
/// @note Copyright &copy; 2013 by Smartware Computing, all rights reserved.
///     Individuals may use this application for evaluation or non-commercial
///     purposes. Within this restriction, changes may be made to this application
///     as long as this copyright notice is retained. The user shall make
///     clear that their work is a derived work, and not the original.
///     Users of this application and sources accept this application "as is" and
///     shall hold harmless Smartware Computing, for any undesired results while
///     using this application - whether real or imagined.
///
/// @author David Smart, Smartware Computing
///
#include "mbed.h"               // ver 63
#include "SDFileSystem.h"       // ver 2, standard

// Modified standard components
#include "WiflyInterface.h"     // ver 4+, derived from version 4 with my mods

// My components
#include "SW_HTTPServer.h"      // ver 0, the initial release.
#include "Utility.h"

Serial pc(USBTX, USBRX);

LocalFileSystem local("local");         // some place to hold settings
SDFileSystem sd(p5, p6, p7, p8, "sd");  // for the static web pages

BusOut leds(LED1,LED2,LED3,LED4);       // dynamic page "/dyn2" lets you control these

// CopyFile
//
// An idea that didn't pan out since the local file does not support folders
//
bool CopyFile(const char * dst, const char * src) {
    FILE * fsrc;
    FILE * fdst;
    char fbuffer[1000];
    int bytes;
    
    pc.printf("Copy to %s from %s\r\n", dst, src);
    fsrc = fopen(src,"rb");
    if (fsrc) {
        pc.printf("  opened %s\r\n", src);
        fdst = fopen(dst,"wb");
        if (fdst) {
            pc.printf("  opened %s\r\n", dst);
            bytes = fread(fbuffer,sizeof(char),sizeof(fbuffer),fsrc);
            while (bytes > 0) {
                fwrite(fbuffer, sizeof(char), bytes, fdst);
                bytes = fread(fbuffer,sizeof(char),sizeof(fbuffer),fsrc);
            }
            fclose(fdst);
            fclose(fsrc);
            pc.printf("  success.\r\n");
            return true;
        }
        fclose(fsrc);
    }
    pc.printf("  failed.\r\n");
    return false;
}

/// SimplyDynamicPage
///
/// This web page is generated dynamically and fills in part of the information from 
/// data extracted from the URL.
///
/// You can see in main how this page was registered.
///
void SimpleDynamicPage(HTTPServer *svr, const char * path, const HTTPServer::namevalue *params, int paramcount) {
    char buf[100];

    // send the header
    svr->header(200, "OK", "Content-Type: text/html\r\n");
    // send some data
    svr->send("<html><head><title>Dynamic Page</title></head>\r\n");
    svr->send("<body>\r\n");
    svr->send("This page was generated dynamically. Create your own name=value pairs on the URL "
              "which uses the GET method.<br/>\r\n");
    sprintf(buf, "%d parameters passed to {%s}:<br/>\r\n", paramcount, path);
    svr->send(buf);
    // show each of the parameters passed on the URL
    for (int i=0; i<paramcount; i++) {
        sprintf(buf, "%d: %s = %s<br/>\r\n", i, params[i].name, params[i].value);
        svr->send(buf);
    }
    svr->send("the end.</body></html>\r\n");
}

/// SimplyDynamicForm
///
/// This web page is generated dynamically and fills in part of the information from 
/// data extracted from the URL. It also generates a web form, and allows you to
/// update the data in the form using the POST method.
///
/// You can see in main how this page was registered.
///
void SimpleDynamicForm(HTTPServer *svr, const char * path, const HTTPServer::namevalue *params, int paramcount) {
    char buf[100];
    
    // set the LEDs based on a passed in parameter.
    leds = atoi(svr->GetParameter("leds"));
    // send the header
    svr->header(200, "OK", svr->GetSupportedType(".htm"));  //svr->header(200, "OK", "Content-Type: text/html\r\n");
    // send some data
    svr->send("<html><head><title>Dynamic Page</title></head>\r\n");
    svr->send("<body>\r\n");
    svr->send("This form was generated dynamically. You can add name=value pairs on the URL "
              "which will show up in the form, but the form is submitted using POST method.<br/>\r\n");
    svr->send("Hint: leds=7 turns on 3 of the 4 blue leds on the mbed<br/>\r\n");
    sprintf(buf, "%d parameters passed to {%s}:<br/>\r\n", paramcount, path);
    svr->send(buf);
    // Create a user form for which they can post changes
    sprintf(buf, "<form method='post' action='%s'>\r\n", path);
    svr->send(buf);
    // show the parameters in a nice format
    svr->send("<table>\r\n");
    for (int i=0; i<paramcount; i++) {
        sprintf(buf, "<tr><td>%d</td><td>%s</td><td><input type='text' name='%s' value='%s'></td></tr>\r\n", i, params[i].name, params[i].name, params[i].value);
        svr->send(buf);
    }
    svr->send("<tr><td>&nbsp;</td><td colspan='2'><input type='submit' value='submit'><input type='reset' value='clear'></td></tr>\r\n");
    svr->send("</table>\r\n");
    svr->send("</form>\r\n");
    // see how we're doing with free memory
    sprintf(buf,"Free memory space: %d\r\n", Free());
    svr->send(buf);
    svr->send("the end.</body></html>\r\n");
}

/// This section helps manage the credentials. I don't like the examples
/// that show clear text credentials, even if they are simply the 
/// access point ssid and passphrase. So, I create this to let me
/// edit a simple file on the file system and put them in there.
/// Marginally more secure, but at least not in the source code that
/// I might share with others.
///
typedef struct
{
    char * ssid;
    char * pass;
} credentials;

#define CREDENTIAL_SIZE 50
bool ReadCredentials(char * file, credentials * cr)
{
    FILE *fp = fopen(file, "r");
    
    cr->ssid = (char *)malloc(CREDENTIAL_SIZE);
    cr->pass = (char *)malloc(CREDENTIAL_SIZE);
    if (fp) {
        fgets(cr->ssid, CREDENTIAL_SIZE, fp);
        fgets(cr->pass, CREDENTIAL_SIZE, fp);
        fclose(fp);
        RTrim(cr->ssid);
        RTrim(cr->pass);
        return true;
    } else {
        return false;
    }
}

void FreeCredentials(credentials * cr)
{
    // first secure them by wiping them out
    for (int i=0; i<CREDENTIAL_SIZE; i++) {
        cr->ssid[i] = cr->pass[i] = '*';
    }
    // then free the memory
    free(cr->ssid);
    free(cr->pass);
}


/// OK, let's get down to business here - 
///
int main() {
    credentials cr;     // handles the credentials
    
    pc.baud(460800);    // I like a snappy terminal, so crank it up!
    
    pc.printf("\r\nSmart WiFly 4 - build " __DATE__ " " __TIME__ "\r\n");

    if (!ReadCredentials("/local/user.txt", &cr)) {
        pc.printf("**** ERROR, no /local/user.txt file was found. ****\r\n");
        pc.printf("     awaiting watchdog (that means you push the button).\r\n");
        while(1);
    }
    
    // instantiate the WiflyInterface, then release the credentials
    WiflyInterface wifly(p28, p27, p23, p24, cr.ssid, cr.pass, WPA);
    FreeCredentials(&cr);

    // start it up as a client of my network    
    wifly.init(); // use DHCP
    
    wifly.baud(230400);     // I like this fast, but it drops chars at 460800
    
    while (!wifly.connect()) {
        printf(" ... failed to connect, retrying.\r\n");
        wifly.reset();
        wait(1.0);
    }
    printf("IP Address is %s\n\r", wifly.getIPAddress());

    // Copy /sd/web to /local/web [local file system does not support subdirs]
    // I wanted to first image the sd card to the local file system for easier
    // access, and then reverse it for easier updates, but was not able to do this.
    if (0) {
        printf("Open directory on /sd/web\r\n");
        DIR *d = opendir("/sd/web");
        struct dirent *p;
        while((p = readdir(d)) != NULL) {
            printf("%s\r\n", p->d_name);
            char sfqfn[40], dfqfn[40];
            sprintf(sfqfn,"%s/%s","/sd/web", p->d_name);
            sprintf(dfqfn,"%s/%s","/local", p->d_name);
            CopyFile(dfqfn,sfqfn);
        }
        closedir(d);
        printf("Directory closed\r\n");
    }

    // Now let's instantiate the web server - along with a few settings
    // among which is the file system path to the static pages.
    #define HTTP_SERVER_PORT 80
    HTTPServer svr(&wifly, HTTP_SERVER_PORT, "/sd/web/", 30, 10, &pc);
    
    // But for even more fun, I'm registering a couple of dynamic pages
    // which you saw the handlers for up above. Here you can see the path
    // to place on the URL.
    svr.RegisterHandler("/dyn1", SimpleDynamicPage);
    svr.RegisterHandler("/dyn2", SimpleDynamicForm);
    
    // Let the human know it is ready - if they are watching
    pc.printf("Waiting for a connection...\r\n");
    while (true)
    {
        svr.ip_process();   // unfortunately, this is a blocking process
    }
}

